/*
             LUFA Library
     Copyright (C) Dean Camera, 2015.

  dean [at] fourwalledcubicle [dot] com
           www.lufa-lib.org
*/

/*
  Copyright 2015  Dean Camera (dean [at] fourwalledcubicle [dot] com)

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaims all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/** \file
 *
 *  Main source file for the USBtoSerial project. This file contains the main tasks of
 *  the project and is responsible for the initial application hardware configuration.
 */

#include <stdlib.h>
#include "gpib.h"
#include "USBGPIB.h"

#define BUFSIZE 240
volatile unsigned char buffer[BUFSIZE];

extern unsigned char gp_result;
volatile unsigned int cpt_sof;

/** LUFA CDC Class driver interface configuration and state information. This structure is
 *  passed to all CDC Class driver functions, so that multiple instances of the same class
 *  within a device can be differentiated from one another.
 */
USB_ClassInfo_CDC_Device_t VirtualSerial_CDC_Interface =
  {
    .Config =
    {
      .ControlInterfaceNumber         = INTERFACE_ID_CDC_CCI,
      .DataINEndpoint                 =
      {
	.Address                = CDC_TX_EPADDR,
	.Size                   = CDC_TXRX_EPSIZE,
	.Banks                  = 1,
      },
      .DataOUTEndpoint                =
      {
	.Address                = CDC_RX_EPADDR,
	.Size                   = CDC_TXRX_EPSIZE,
	.Banks                  = 1,
      },
      .NotificationEndpoint           =
      {
	.Address                = CDC_NOTIFICATION_EPADDR,
	.Size                   = CDC_NOTIFICATION_EPSIZE,
	.Banks                  = 1,
      },
    },
  };


/** Standard file stream for the CDC interface when set up, so that the virtual
CDC COM port can be
 *  used like any regular character stream in the C APIs
 */
static FILE USBSerialStream;

/** Main program entry point. This routine contains the overall program flow, including initial
 *  setup of all components and the main program loop.
 */
int main(void)
{
  SetupHardware();

  /* Create a regular character stream for the interface so that it can be
     used with the stdio.h functions */
  CDC_Device_CreateStream(&VirtualSerial_CDC_Interface, &USBSerialStream);
 
  GlobalInterruptEnable();

  for (;;)
    {
      if( CDC_Device_BytesReceived(&VirtualSerial_CDC_Interface) > 0 ){
        unsigned char pos = 0;
        while (1){
          if( CDC_Device_BytesReceived(&VirtualSerial_CDC_Interface) > 0 ){
            unsigned char ch = fgetc(&USBSerialStream);
            if(ch == '\n' || ch == '\0'){
              break;
            } else if( ch >= 0x20 && ch <= 0x7F ) { //command line is ASCII
              buffer[pos++] = ch;
              if(pos==BUFSIZE-1)
                break;
            }
          }
        }
        buffer[pos] = '\0';
        char comm,comm2;
        unsigned char addr,pb,pd,eoi;
        char *dat;
        unsigned char i;
        long size;
        if(strncmp((const char *)buffer,"REN",3)==0){
          gp_set_ren(1);
        } else if (strncmp((const char *)buffer,"RDI",3)==0){
          gp_set_ren(0);
        } else if (strncmp((const char *)buffer,"IFC",3)==0){
          gp_set_ifc();
        } else if (strncmp((const char *)buffer,"STA",3)==0){
          gp_line_status( &pb, &pd );
          fprintf(&USBSerialStream,"PB = %02x,  PD = %02x\r\n",pb,pd);
        } else if (strncmp((const char *)buffer,"SRQ",3)==0){
	  buffer[0] = gp_srq();
	  CDC_Device_SendData(&VirtualSerial_CDC_Interface, buffer, 1);	  
        } else {
          comm = buffer[0];
          addr = (buffer[1]-48) * 10 + buffer[2] - 48;
          dat = (char *)&buffer[3];
          switch(comm){
          case 'C' : // report cpt_sof, for check alive
            fprintf(&USBSerialStream, "cpt_sof = %d\r\n", cpt_sof );
            break;
          case 'D' : // set delimiter
            gp_set_delim( addr, *dat - 48 );
            break;
          case 'E' : // get gp_result
	    CDC_Device_SendData(&VirtualSerial_CDC_Interface, &gp_result, 1);
            break;
          case 'S' : // send ASCII string to addr
            gp_sendline( addr, (char *)dat, strlen(dat) );
            break;
          case 'R' : // receive ASCII string from addr
            gp_readline( addr, (char *)buffer, BUFSIZE );
            fprintf(&USBSerialStream, "%s\r\n", buffer );
            break;
	  case 'M' : // multiline message
            // 0x20 is added to the command byte to express it in ASCII
            gp_multiline( addr, dat[0] - 0x20 );
            break;
          case 'P' : // serial poll
	    buffer[0] = gp_spoll(addr);
	    CDC_Device_SendData(&VirtualSerial_CDC_Interface, buffer, 1);
            break;
          case 'B' : //binary transfer
            comm2 = buffer[1];
            addr = (buffer[2]-48) * 10 + buffer[3] - 48;
            size = atol((char *)&buffer[4]);
            switch(comm2){
            case 'S' : //send binary array
              gp_set_listener( addr );
              //gp_set_output(); No need here
              while(size){
                if(CDC_Device_BytesReceived(&VirtualSerial_CDC_Interface)>0){
                  gp_wbyte( fgetc(&USBSerialStream), size==1 );
                  size--;
                }
              }
              break;
            case 'R' : //receive binary array
              gp_set_talker( addr );
              gp_set_input();
              eoi = 1;
              while(eoi && size){
                i=0;
                // transfer several times if the data length
                // is greater than the buffer size
                while(eoi && size && (i<BUFSIZE)){
                  buffer[i] = gp_rbyte( &eoi );
                  if(gp_result) break;
                  i++;
		  size--;
                }
                if(i) CDC_Device_SendData(&VirtualSerial_CDC_Interface, buffer, i);
		if (gp_result) break; // avoid eternal loop
              }
              if(size>0) gp_result = 0xFF; // too short
              break;
            }
          }
        }
      }

      CDC_Device_USBTask(&VirtualSerial_CDC_Interface);
      USB_USBTask();
    }
}

/** Configures the board hardware and chip peripherals for the demo's functionality. */
void SetupHardware(void)
{
#if (ARCH == ARCH_AVR8)
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();
  
  /* Disable clock division */
  clock_prescale_set(clock_div_1);
#endif

  gp_init();
  USB_Init();

  USB_Device_EnableSOFEvents();
}

/** Event handler for the library USB Connection event. */
void EVENT_USB_Device_Connect(void)
{
}

/** Event handler for the library USB Disconnection event. */
void EVENT_USB_Device_Disconnect(void)
{
}

/** Event handler for the library USB Configuration Changed event. */
void EVENT_USB_Device_ConfigurationChanged(void)
{
  bool ConfigSuccess = true;

  ConfigSuccess &= CDC_Device_ConfigureEndpoints(&VirtualSerial_CDC_Interface);
}

/** Event handler for the library USB Control Request reception event. */
void EVENT_USB_Device_ControlRequest(void)
{
  CDC_Device_ProcessControlRequest(&VirtualSerial_CDC_Interface);
}

/** CDC class driver callback function the processing of changes to the virtual
 *  control lines sent from the host..
 *
 *  \param[in] CDCInterfaceInfo  Pointer to the CDC class interface configuration structure being referenced
 */
void EVENT_CDC_Device_ControLineStateChanged(USB_ClassInfo_CDC_Device_t *const CDCInterfaceInfo)
{
  /* You can get changes to the virtual CDC lines in this callback; a common
     use-case is to use the Data Terminal Ready (DTR) flag to enable and
     disable CDC communications in your application when set to avoid the
     application blocking while waiting for a host to become ready and read
     in the pending data from the USB endpoints.
  */
  bool HostReady = (CDCInterfaceInfo->State.ControlLineStates.HostToDevice & CDC_CONTROL_LINE_OUT_DTR) != 0;
}

/* this routine is expected to be called every 1ms */
void EVENT_USB_Device_StartOfFrame(void)
{
  cpt_sof++;
}

