#include <avr/io.h>
#include <avr/eeprom.h>
//#define F_CPU 16000000
#include <util/delay.h>
#include "gpib.h"

#include <stdio.h>

/* GPIB timeout in ms */
#define TIMEOUT 500

/* GPIB delay after ATN in us */
#define ATN_WAIT 5

#define REN  0x01
#define EOI  0x02
#define DAV  0x04
#define NRFD 0x08
#define NDAC 0x10
#define IFC  0x20
#define SRQ  0x40
#define ATN  0x80

#define SET_OUTPUT(pin) (DDRB |= (pin))
#define SET_INPUT(pin) (DDRB &= ~(pin))
#define SET_HIGH(pin) (PORTB |= (pin))
#define SET_LOW(pin) (PORTB &= ~(pin))

extern volatile unsigned int cpt_sof;
unsigned char gp_result;

void gp_init(void)
{
  SET_OUTPUT( REN | EOI | DAV | NRFD | NDAC | IFC | ATN );
  SET_INPUT( SRQ ); //SRQ̂ݓ
  DDRD = 0xFF; // all INPUT

  SET_HIGH( REN | EOI | DAV | IFC | ATN );
  SET_LOW( NRFD | NDAC );
  _delay_ms(1);

  gp_result = 0;
}
//
// set delimiter for ASCII transfer
//
void gp_set_delim( char addr, char dm )
{
// dm = 0:EOI only, 1:EOI or CR, 2:EOI or LF, 3:EOI or CRLF
  eeprom_write_byte( addr, dm );
  gp_result = 0;
}
//
// configure for input 
//
void gp_set_input(void)
{
  SET_OUTPUT( NRFD | NDAC );
  SET_INPUT( DAV | EOI );
  DDRD = 0x00;
  SET_LOW( NRFD | NDAC );
  gp_result = 0;
}
//
// configure for output
//
void gp_set_output(void)
{
  SET_INPUT( NRFD | NDAC );
  SET_OUTPUT( DAV | EOI );
  DDRD = 0xFF;
  SET_HIGH( EOI | DAV );
  gp_result = 0;
}
//
// inquire line status
//
void gp_line_status(unsigned char *pb, unsigned char *pd)
{
  SET_INPUT( EOI | DAV | NRFD | NDAC );
  DDRD = 0; // All INPUT
  _delay_ms(1);
  *pd = ~PIND;
  *pb = ~PINB;

  SET_OUTPUT( EOI | DAV | NRFD | NDAC );
  DDRD = 0xFF; // All OUTPUT
  SET_HIGH( EOI | DAV );
  SET_LOW( NRFD | NDAC );
  gp_result = 0;
}
//
// set REN
//
void gp_set_ren( char on )
{
  if(on)
    SET_LOW( REN );
  else
    SET_HIGH( REN );

  gp_result = 0;
}
//
// inject IFC pulse
//
void gp_set_ifc(void)
{
  SET_LOW( IFC );
  _delay_us(200);
  SET_HIGH( IFC );

  gp_result = 0;
}
//
// send one byte
//
void gp_wbyte( unsigned char b, char eoi )
{
  cpt_sof = 0;
  while( PINB & NDAC ){ // ensure that NDAC = L
    if( cpt_sof > TIMEOUT ){
      gp_result = 1;
      return;
    }
  }
  if (eoi) SET_LOW( EOI );
  PORTD = ~b; // set data
  while( (PINB & NRFD) == 0 ){ // wait for NRFD = H
    if( cpt_sof > TIMEOUT ){
      gp_result = 2;
      return;
    }
  }
  SET_LOW( DAV );
  while( (PINB & NDAC) == 0 ){ // wait for NDAC = H
    if( cpt_sof > TIMEOUT ){
      gp_result = 3;
      return;
    }
  }
  SET_HIGH( DAV );
  if (eoi) SET_HIGH( EOI );
  PORTD = 0xFF; // clear data
  gp_result = 0;
}
//
// receive one byte
//
unsigned char gp_rbyte( unsigned char *eoi_f )
{
  unsigned char b;

  cpt_sof = 0;
  SET_HIGH( NRFD );
  while( PINB & DAV ){ // wait for DAV = L
    if( cpt_sof > TIMEOUT ){
      gp_result = 4;
      return 0;
    }
  }
  SET_LOW( NRFD );
  b = ~PIND;
  *eoi_f = (PINB & EOI);
  SET_HIGH( NDAC );
  while((PINB & DAV) == 0) { // wait for DAV = H 
    if( cpt_sof > TIMEOUT ){
      gp_result = 5;
      return 0;
    }
  }
  SET_LOW( NDAC );

  gp_result = 0;
  return b;
}
//
// set listener
//
void gp_set_listener( char addr )
{
  gp_set_output();

  SET_LOW( ATN );
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
  gp_wbyte( 0x3F, 0 ); //Unlisten
  if(gp_result!=0) { gp_result |= 8; return; }
  gp_wbyte( 0x40, 0 ); //Talker is me
  if(gp_result!=0) { gp_result |= 8; return; }
  gp_wbyte( 0x20 + addr, 0 ); //Listener  
  if(gp_result!=0) { gp_result |= 8; return; }
  SET_HIGH( ATN ); 
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
}
//
// set talker
//
void gp_set_talker(char addr)
{
  gp_set_output();

  SET_LOW( ATN );
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
  gp_wbyte( 0x3F, 0 ); //Unlisten
  if(gp_result!=0) { gp_result |= 16; return; }
  gp_wbyte( 0x40 + addr, 0 ); //Talker
  if(gp_result!=0) { gp_result |= 16; return; }
  gp_wbyte( 0x20, 0 ); //Listener is me
  if(gp_result!=0) { gp_result |= 16; return; }
  SET_HIGH( ATN );
#ifdef ATN_WAIT
  _delay_us( ATN_WAIT );
#endif
}
//
// send one line to the gpib line
//
void gp_sendline( char addr, char *data, int len )
{
  char delim = eeprom_read_byte( addr );

  if(delim & 1) data[len++] = '\r'; // add CR
  if(delim & 2) data[len++] = '\n'; // add LF

  gp_set_listener( addr );
  if(gp_result!=0) { gp_result |= 32; return; }
  // gp_set_output(); No need here

  while(len){
    gp_wbyte( *data++, len==1 ); // len==1 : final byte
    if(gp_result != 0) { 
      gp_result |= 32; 
      return; 
    }
    len--;
  }
}
//
// receive one line from gpib line
//
void gp_readline(char addr, char *buf, int len )
{
  buf[0] = '\0';

  // for test
  //sprintf(buf,"cpt_sof = %d", cpt_sof);
  //return;

  gp_set_talker( addr );
  if(gp_result != 0) { gp_result |= 64; return; }
  gp_set_input();

  unsigned char eoi_f = EOI;
  char del = eeprom_read_byte(addr);
  char flag = 0;
  char ch = 0;
  int i = 0;
  while(i < len && eoi_f){
    ch = gp_rbyte( &eoi_f );
    if(gp_result != 0){
      buf[i] = '\0';
      gp_result |= 64;
      return;
    }
    if (ch=='\r') {
      flag |= 1;
      if( del && del==flag ) break;
    } else if (ch=='\n') {
      flag |= 2;
      if( del && del==flag ) break;
    } else {
      buf[i++] = ch;
      flag = 0;
    }
  }
  buf[i] = '\0';
}

void gp_multiline( char addr, char comm )
{
  gp_set_output();

  SET_LOW( ATN );
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
  if(addr!=0){
    gp_wbyte( 0x3F, 0 ); //Unlisten
    if(gp_result!=0) {gp_result |= 128; return;}
    gp_wbyte( 0x20 + addr, 0 ); //Listener
    if(gp_result!=0) {gp_result |= 128; return;}
  }
  gp_wbyte( comm, 0 ); //Multibyte Command
  if(gp_result!=0) {gp_result |= 128; return; }
  SET_HIGH( ATN );
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
  gp_result = 0;
}

char gp_srq(void)
{
  gp_result = 0;
  return ((PINB & SRQ)==0);
}

unsigned char gp_spoll(char addr)
{
  unsigned char stb;
  unsigned char eoi_f;

  gp_set_output();
  SET_LOW( ATN );
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
  gp_wbyte( 0x3F, 0 ); //Unlisten
  if(gp_result!=0) return 0xFF;
  gp_wbyte( 0x18, 0 ); //SPE
  if(gp_result!=0) return 0xFF;
  gp_wbyte( 0x40 + addr, 0 ); //Talker
  if(gp_result!=0) return 0xFF;
  SET_HIGH( ATN );
  gp_set_input();
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
  stb = gp_rbyte(&eoi_f); // receive status byte
  if(gp_result!=0) return 0xFF;
  gp_set_output();
  SET_LOW( ATN );
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
  gp_wbyte( 0x19, 0 ); //SPD
  if(gp_result!=0) return 0xFF;
  gp_wbyte( 0x5F, 0 ); //UNT
  if(gp_result!=0) return 0xFF;
  SET_HIGH( ATN );
#ifdef ATN_WAIT
  _delay_us(ATN_WAIT);
#endif
  return stb;  
}
